/*
 * Copyright (C) 2006 Apple Computer, Inc.  All rights reserved.
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE COMPUTER, INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 */

#include "config.h"
#include "ImageDecoder.h"
#include "NotImplemented.h"
#include "GraphicsTypes.h"
#include "GraphicsContextPlatformPrivate.h"
#include "StatisticsWiiU.h"
#include "ImageDecoderWiiU.h"
#include <PlatformOS.h>

using namespace WTF;
namespace WebCore {
extern "C" {
	extern int gForce16bitAlphaTextures;
	extern int gForce16bitOpaqueTextures;
	extern int gForceBC1Textures;
};


// static helper functions
void* ImageWiiU::createTexture(int width, int height, bool hasAlpha, void *sourcePixels, bool allowBC1, bool allowFormats)
{
	void* texture;
	GraphicsContextPlatformPrivate::TextureFormat fmt = GraphicsContextPlatformPrivate::TEX_FMT_RGBA8;
	m_dither = false;
	if(allowFormats)
	{
		if(hasAlpha)
		{
			if(gForce16bitAlphaTextures)
			{
				fmt = GraphicsContextPlatformPrivate::TEX_FMT_RGBA4444;
				if(gForce16bitAlphaTextures > 1)
					m_dither = true;
			}
		}
		else
		{
			if(gForceBC1Textures && allowBC1)	// only opaque textures can be BC1
				fmt = GraphicsContextPlatformPrivate::TEX_FMT_BC1;
			else if(gForce16bitOpaqueTextures)
			{
				fmt = GraphicsContextPlatformPrivate::TEX_FMT_RGBA5551;
				if(gForce16bitOpaqueTextures > 1)
					m_dither = true;
			}
		}
	}
	m_format = fmt;
	if(m_format == GraphicsContextPlatformPrivate::TEX_FMT_BC1)	
	// BC1 now requires two textures for a little while. 
	// for images that are rendered as they are downloaded from the internet, constant recompression is extremely slow, and with the intermediate image buffer removed
	// to decompress directly to hardware textures, an uncompressed format will be used to render the texture until the image is done being decompressed (signalled via the
	// "complete" member function
	{
		// create the secondary texture for BC1 here
		m_secondTextureP = GraphicsContextPlatformPrivate::createTextureFromMemoryOnly(width, height, GraphicsContextPlatformPrivate::TEX_FMT_BC1);
		if(m_secondTextureP == NULL)
			return NULL;	// failed, no texture allowed if BC1 version can't be allocated.
		
		// special case here, set "fmt" to RGBA32 bit so that the main texture is RGBA8 until conversion is complete
		fmt = GraphicsContextPlatformPrivate::TEX_FMT_RGBA8;
	}
	texture = GraphicsContextPlatformPrivate::createTextureFromMemoryOnly(width, height, fmt);
	return texture;
}

// new class for native images to share between decoder and native
ImageWiiU* ImageWiiU::create(int width, int height, bool hasAlpha, void *sourcePixels, bool allowBC1, bool allowFormats)
{
	ImageWiiU *imageP = new ImageWiiU(width, height, hasAlpha, sourcePixels, allowBC1, allowFormats);
	return imageP;
}

ImageWiiU* ImageWiiU::create(void *textureDataP, bool isDirect)
{
	return new ImageWiiU(textureDataP, isDirect);
}

ImageWiiU* ImageWiiU::create(ImageWiiU *imageP)
{
	ImageWiiU *imageWiiuP = new ImageWiiU(imageP);
	if(imageWiiuP)
	{
		if(imageWiiuP->m_textureP == NULL && imageP->m_textureP != NULL)	// big failure to create the copy
		{
			delete imageWiiuP;
			imageWiiuP = NULL;
		}
	}
	return imageWiiuP;
}

ImageWiiU::ImageWiiU(int width, int height, bool hasAlpha, void *sourcePixels, bool allowBC1, bool allowFormats)
	: m_refCount(1)
	, m_textureP(NULL)
	, m_secondTextureP(NULL)
	, m_format(GraphicsContextPlatformPrivate::TEX_FMT_RGBA8)
	, m_dither(false)
	, m_isDirect(false)
	, m_isDirty(false)
	, m_isFinished(false)
{
	m_textureP = createTexture(width, height, hasAlpha, sourcePixels, allowBC1, allowFormats);
}

ImageWiiU::ImageWiiU(void *textureDataP, bool isDirect)
	: m_refCount(1)
	, m_textureP(textureDataP)
	, m_secondTextureP(NULL)
	, m_format(GraphicsContextPlatformPrivate::TEX_FMT_RGBA8)
	, m_dither(false)
	, m_isDirect(isDirect)
	, m_isDirty(false)
	, m_isFinished(true)
{
	// nothing else is done except for member initialization
}

ImageWiiU::ImageWiiU(ImageWiiU *originalP)
	: m_refCount(1)
	, m_format(originalP->m_format)
	, m_dither(originalP->m_dither)
	, m_isDirect(originalP->m_isDirect)
	, m_isDirty(true)
	, m_isFinished(false)
{
	// copy the data over exactly?
	if(originalP->m_textureP)
		m_textureP = GraphicsContextPlatformPrivate::copyTexture(originalP->m_textureP);
	else
		m_textureP = NULL;
	if(originalP->m_secondTextureP)
		m_secondTextureP = GraphicsContextPlatformPrivate::copyTexture(originalP->m_secondTextureP);
	else
		m_secondTextureP = NULL;
}

ImageWiiU::~ImageWiiU(void)
{
	if(m_isDirect == false)
	{
		if(m_textureP)
			WKgfx::DelayedNativeImageFree(m_textureP);
		m_textureP = NULL;
	}
}

void ImageWiiU::ref(void)
{
	m_refCount++;
}

void ImageWiiU::unref(void)
{
	m_refCount--;
	if(m_refCount <= 0)
		delete this;
}


void ImageWiiU::writeNativeImagePixelsOpaque(int y, int width, unsigned char *sourcePixelsP, int step)
{
	if(m_textureP)
	{
		GraphicsContextPlatformPrivate::writeTextureLineOpaque(m_textureP, y, width, sourcePixelsP, step, m_dither);
		m_isDirty = true;
	}
}

bool ImageWiiU::writeNativeImagePixelsAlpha(int y, int width, unsigned char *sourcePixelsP, int step, bool preMultiplyAlpha)
{
	if(m_textureP)
	{
		m_isDirty = true;
		return GraphicsContextPlatformPrivate::writeTextureLineAlpha(m_textureP, y, width, sourcePixelsP, preMultiplyAlpha, m_dither);
	}
	return false;
}

void ImageWiiU::writeNativeImagePixels(int y, int width, unsigned long *sourcePixelsP)
{
	if(m_textureP)
	{
		m_isDirty = true;
		GraphicsContextPlatformPrivate::writeTextureLine(m_textureP, y, width, sourcePixelsP, m_dither);
	}
}

bool ImageWiiU::writeNativeImagePixel(int x, int y, unsigned long argb)
{
	if(m_textureP)
	{
		m_isDirty = true;
		GraphicsContextPlatformPrivate::writeTexturePixel(m_textureP, x, y, argb, m_dither);
		return true;
	}
	return false;
}

bool ImageWiiU::writeNativeImagePixel(int x, int y, unsigned char r, unsigned char g, unsigned char b, unsigned char a, bool preMultiplyAlpha)
{
	if(m_textureP)
	{
		bool retVal = false;
        if (preMultiplyAlpha && a < 255) {
            if (!a) {
                r = g = b = 0;
            }
			else
			{
				unsigned alphaMult = a * ImageFrame::fixPointMult;
				r = ImageFrame::fixPointUnsignedMultiply(r, alphaMult);
				g = ImageFrame::fixPointUnsignedMultiply(g, alphaMult);
				b = ImageFrame::fixPointUnsignedMultiply(b, alphaMult);
				retVal = true;
			}
        }
		m_isDirty = true;
		GraphicsContextPlatformPrivate::writeTexturePixel(m_textureP, x, y, ((unsigned int)a << 24 | (unsigned int)r << 16 | (unsigned int)g << 8 | (unsigned int)b), m_dither);
		return retVal;
	}
	return false;
}

void ImageWiiU::copyRowNTimes(int startX, int endX, int startY, int endY)
{
	if(m_textureP)
		GraphicsContextPlatformPrivate::copyTextureRowNTimes(m_textureP, startX, endX, startY, endY);
	m_isDirty = true;
}

void ImageWiiU::zeroData(void)
{
	if(m_textureP)
	{
		GraphicsContextPlatformPrivate::clearTexture(m_textureP);
		m_isDirty = true;
	}
}

void ImageWiiU::invalidateData(void)
{
	if(m_isDirty)
	{
		if(m_textureP)
		{
			if(m_isFinished == false)
			{
				GraphicsContextPlatformPrivate::finishTextureConversion(m_textureP);
				if(m_secondTextureP) // if the secondary texture exists, it must be finished as well
					GraphicsContextPlatformPrivate::finishTextureConversion(m_secondTextureP, false);
				m_isFinished = true;
			}
			else
				GraphicsContextPlatformPrivate::invalidateTexture(m_textureP);
			m_isDirty = false;
		}
	}
}

void ImageWiiU::complete(void)
{
	// this new function is specifically used for BC1 texture conversion, and is called when an image decoder says the frame is "complete" (ie: decoding finished)
	// this will convert the RGBA8888 texture used during display while downloading as a source, convert it to BC1, then schedule a deletion of the RGBA8888 texture
	// and reassign the BC1 as the new "main" texture
	if(m_secondTextureP)	// convert to BC1, finally
	{
		if(m_textureP && m_format == GraphicsContextPlatformPrivate::TEX_FMT_BC1)	// sanity check; should already exist
		{
			GraphicsContextPlatformPrivate::convertTextureFromMemoryToNativeBC1(m_secondTextureP, m_textureP);	// internally this handles conversion process
			// in case the BC1 texture did not have its internal setup finished prior, it will be done now
			if(m_secondTextureP && m_isFinished == false)
				GraphicsContextPlatformPrivate::finishTextureConversion(m_secondTextureP, false);
			WKgfx::DelayedNativeImageFree(m_textureP);
			m_textureP = m_secondTextureP;
			m_isDirty = false;
			m_isFinished = true;
		}
		else
		{
			OS::Report("Error - format (%ld) and texture (0x%08x) not in place for BC1 conversion\n", m_format, m_textureP);
		}
		m_secondTextureP = NULL;
	}
}

void* ImageWiiU::getTexture(void)
{
	invalidateData();
	return m_textureP;
}

// ****************************************************************************************
// this is the WiiU Image Frame function location
// ****************************************************************************************
void ImageFrame::copyRowNTimes(int startX, int endX, int startY, int endY)
{
	if(m_imageP)
		m_imageP->copyRowNTimes(startX, endX, startY, endY);
}

void ImageFrame::setDirectTexture(void *textureDataP)
{
	if(m_imageP)
	{
		m_imageP->unref();
		m_imageP = NULL;
	}
	m_imageP = ImageWiiU::create(textureDataP);
}

NativeImagePtr ImageFrame::asNewNativeImage(void) const
{
	if(m_imageP)
	{
		m_imageP->ref();
		STATS_WIIU_INC_IMAGE_COUNT();
		return m_imageP;
	}
    return NULL;
}

} // namespace WebCore
