/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef JSEXTENSIONUTILS_H
#define JSEXTENSIONUTILS_H

#include <string>
#include <map>
#include <vector>

#include <WebCore/bindings/js/JSDOMGlobalObject.h>
#include <WebCore/html/HTMLImageElement.h>
#include <WebCore/fileapi/FileReader.h>
#include <WebCore/loader/cache/CachedImage.h>
#include <WebCore/platform/graphics/Image.h>
#include <WebCore/platform/graphics/IntSize.h>
#include <WebCore/platform/image-encoders/JPEGImageEncoder.h>
#include <WebCore/platform/network/BlobData.h>
#include <WebCore/platform/network/BlobStorageData.h>
#include <WebCore/platform/network/BlobRegistry.h>
#include <WebCore/platform/network/BlobRegistryImpl.h>

#include <JavaScriptCore/API/JavaScript.h>

#include <DerivedSources/WebCore/JSBlob.h>
#include <DerivedSources/WebCore/JSHTMLImageElement.h>

#include "JSEventSystem.h"

// If you declare your static functions/values with the following macros,
// Then you can make use of the CHECK_SCOPE and CHECK_ARGS macros.
//
// Example)	JSValueRef createIcon(JSFUNCTIONARGS) {
//			CHECK_ENTRY(class_reference, 2, 1, kJSTypeObject, kJSTypeObject);
//			...
//
#define JSFUNCTIONARGS \
	JSContextRef context, \
	JSObjectRef function, \
	JSObjectRef object_reference, \
	size_t argumentCount, \
	const JSValueRef arguments[], \
	JSValueRef *exception

#define JSFUNCTIONARGS_NAMES \
	context, \
	function, \
	object_reference, \
	argumentCount, \
	arguments, \
	exception

#define JSPROPERTYARGS \
	JSContextRef context, \
	JSObjectRef object_reference, \
	JSStringRef propertyName, \
	JSValueRef *exception

#define JSSETPROPERTYARGS \
	JSContextRef context, \
	JSObjectRef object_reference, \
	JSStringRef propertyName, \
	JSValueRef value, \
	JSValueRef *exception

#define CHECK_SCOPE(class_reference) \
	if(!JSExtUtils::checkCallbackScope(context, object_reference, class_reference, exception)) { \
		if(exception) return *exception; \
		return JSValueMakeUndefined(context); \
	}

#define CHECK_ARGS(max_args, num_default_args, ...) \
	JSType expected_types[] = {__VA_ARGS__}; \
	if(!JSExtUtils::checkArgumentList(max_args, expected_types, context, argumentCount, arguments, NULL, exception, num_default_args)) { \
		if(exception) return *exception; \
		return JSValueMakeUndefined(context); \
	}

#define CHECK_ENTRY(class_reference, max_args, num_default_args, ...) \
	CHECK_SCOPE(class_reference); \
	CHECK_ARGS(max_args, num_default_args, __VA_ARGS__);

namespace JSExtUtils
{
    void UStringToStdString(const JSC::UString *ustring, std::string *stdString);
    std::string getJSTypeAsString(JSType type);
	bool checkCallbackScope(JSContextRef callbackContext, JSValueRef callbackScope, JSClassRef expectedClass, JSValueRef* exception = NULL);
	bool checkArgumentList(size_t expectedArgumentCount, JSType expectedTypes[], JSContextRef context,  size_t providedArgumentCount, const JSValueRef arguments[], JSObjectRef thisObject, JSValueRef* exception = NULL, size_t numDefaultArguments = 0);
	JSObjectRef getPackageObject(JSContextRef ctx, JSObjectRef base, std::string packageName, JSPropertyAttributes attributes = kJSPropertyAttributeNone);
	bool deleteJSObjectRefByString(JSContextRef ctx, JSObjectRef base, std::string targetName);
	JSObjectRef createErrorObjectWithUTF8String(JSContextRef context, char* errorString, JSValueRef* exception);
	JSValueRef JSObjectGetPropertyByUTF8CString(JSContextRef context, JSObjectRef base, char* targetName, JSValueRef* exception);
	JSValueRef JSObjectListProperties(JSContextRef context, JSObjectRef base, char* objectName);

	//Blob manipulation helper functions
	bool BlobToCharArray(WebCore::Blob* targetBlob, char** outputArray, unsigned* outputArraySize, bool addNullTerminator = false);
	JSObjectRef DataToJSBlobRef(JSContextRef context, char* data, size_t dataSize, JSStringRef jsContentType = 0);
	WebCore::Blob* JSObjectRefToBlob(JSContextRef context, JSValueRef blobObjectRef);

	//Image manipulation helper functions
	void* JSObjectRefToNativeImagePointer(JSContextRef context, JSValueRef imageObjectRef);
	bool NativeImageToJPEGAsCharArray(unsigned char* rgbaBigEndianData, unsigned width, unsigned height, unsigned pitch, char** outputArray, unsigned* outputArraySize, WebCore::ColorByteIndices indices, int quality = 100);
	JSValueRef NativeTextureBundleToImageArray(JSContextRef context, const void *textureBundle, bool bundleIsAligned);

	//Static constant creation functions
	JSValueRef JSObjectSetReadOnlyNumberProperty(JSContextRef context, JSObjectRef baseObject, char* propertyName, double propertyValue);
	JSValueRef JSObjectSetReadOnlyUTF8CStringProperty(JSContextRef context, JSObjectRef baseObject, char* propertyName, char* propertyValue);

	//EventDispatcher helper functions
	void JSObjectAddGenericEventData(JSObjectRef controller, JSContextRef eventContext, JSObjectRef eventClassObj, JSEvent js_event);
	void JSObjectDeleteGenericEventData(JSContextRef eventContext, JSObjectRef eventClassObj);
	JSValueRef GenericAddEventListener(JSFUNCTIONARGS);
	JSValueRef GenericRemoveEventListener(JSFUNCTIONARGS);
	JSValueRef GenericRemoveAllEventListeners(JSFUNCTIONARGS);
	JSValueRef GenericDispatchEvent(JSFUNCTIONARGS);
	JSValueRef GenericHasEventListener(JSFUNCTIONARGS);

    const u8 LANGUAGE_SYSTEM = 200;
}

class JSCustomEventDispatcher : public JSEventDispatcher {
public:

	struct ListenerEventPair {
		ListenerEventPair(void) : js_listener(0) {}
		ListenerEventPair(JSEventListener *js_listener, std::string event_type) : js_listener(js_listener), event_type(event_type) {}
		JSEventListener *js_listener;
		std::string event_type;
	};
	
	JSCustomEventDispatcher(void);
	virtual ~JSCustomEventDispatcher(void);

	virtual bool AddEventListener(JSEventListener *js_listener, std::string event_type);

	virtual void RemoveEventListener(JSEventListener *js_listener, std::string event_type);

	virtual void RemoveAllListeners(JSEventListener *js_listener);

	virtual void DispatchEvent(JSEvent js_event);

	virtual void DispatchEvent(JSEvent js_event, std::string event_string);

private:
	bool isDispatching;
	void processQueuedRequests();

	std::vector<ListenerEventPair> remove_list;
	std::vector<ListenerEventPair> add_list;
	std::vector<JSEventListener *> clear_list;
};

#endif
